/*
 * Voici un extrait du fichier elementDefinition du projet Graphe qui reprend l'héritage et la "notion de package" tel que je l'ai expliqué.
 * J'ai enlevé beaucoup de code que j'avais mis, notament à la déclaration des variables GNode et GEdge, qui était présent pour facilité à l'utilisation des objets dans un IDE.
 * En effet, le code ci-dessous fonctionne parfaitement mais les IDEs ne sont pas capables de projeter l'éxécution de ce code et de déterminer les fonctions et propriétées des différents objets.
 * J'ai laissé les commentaires pour ce qui souhaite investiguer ces éléments et l'utilisation de l'héritage qui est faîte ici.
 * 
 * Nomenclature : les propriétés priver commence par un '_'.
 * 
 * Pour pouvoir se retrouver entre les différentes "propriétés" et "attributs" des objets, j'ai fait des "énumérations".
 * Les "énumérations" en JavaScript sont des objets (tableaux associatif) exemple - enumeration_nom_prenom = {"nom":"nom", "prenom":"prenom"};
 *      node_a = node attribut
 *      node_p = node property
 *      edge_a, edge_p = idem pour les edge
 *      utilisation = correspond à l'utilisation faite des éléments (no-group/origin - groupbyimpact - groupbycause).
 *      debug = une variable pour passé en mode débug et avoir des précisions à destinations des développeur (dans la console ainsi que dans l'UI)
 */

var GNode, GEdge;

var getElements = (function () {
    'use strict';
    /**
     * @description A basic Element.
     * @param {object} data
     * @constructor
     * @private
     */
    function Element (data) {
        /**
         * @description This Element.
         * @type {Element}
         * @private
         */
        var _self = this; // Comment-ajout: cette variable peut sembler inutile car this restera this quoi qu'il se passe.
        // Faut: this est le this trouvé au moment de l'éxécution de la fonction.
        // Une fonction utilisé par une instance qui utilise le this de l'instance, si cette fonction est appelé en asynchrone, le this utilisé n'est plus nécessairement l'instance.
        // Il faut donc faire atttention : Si la fonction est appelé lors d'éxécution de code bloquant qui nécessite d'être terminé avant que le programme ne continu on peut dire que le this sera celui de l'instance.
        // En cas de doute, TOUJOURS privilégier l'utilisation d'une variable qui va conserver la valeur de l'instance comme ici.
        /**
         * @description The attributes of this Element.
         * @type {object}
         * @private
         */
        var _attributes = data; // Comment-ajout: A voir comme les propriétées donné à CytoscapeJS.
        /**
         * @description The properties of this Element.
         * @type {object}
         * @private
         */
        var _properties = {}; // Comment-ajout: A voir comme les "dev properties".
        /**
         * @description The css classes of this Element.
         * @type {string[]}
         */
        var _classes = [];

        /**
         * @description Setter/Getter of data. To get a specific data give a 'dataName', to set a specific data give also her 'dataValue'. To get all data, give nothing.
         * @param {string} [dataName] The name of the data to Get/Set.
         * @param {*} [dataValue] The value of the data to Set.
         * @returns {*} The 'dataValue' if using as a setter, the value of the specified data or all data.
         */
        this.data = function (dataName, dataValue) {
            if (dataValue !== undefined) {
                _attributes.data[dataName] = dataValue;
                return dataValue;
            } else if (dataName !== undefined) {
                return _attributes.data[dataName];
            } else {
                return _attributes.data;
            }
        };
        /**
         * @description Give the 'id' of this Element.
         * @returns {*}
         */
        this.id = function () { return _attributes.data.id; };
        /**
         * @description Return the basic Element (only his attributes).
         * @returns {object}
         */
        this.basicElement = function () { return _attributes; };

        /**
         * @description Set a specific attribute.
         * @param {string} attrName The name of the attribute to set.
         * @param {*} attrValue The value to set to the attribute.
         * @returns {*} The attribute value.
         */
        this.setAttr = function (attrName, attrValue) {
            _attributes[attrName] = attrValue;
            return attrValue;
        };
        /**
         * @description Get a the value of a specific attribute.
         * @param {string} attrName The name of the attribute to get.
         * @returns {*}
         */
        this.getAttr = function (attrName) { return _attributes[attrName]; };
        /**
         * @description Remove a specific attribute.
         * @param {string} attrName The name of the attribute to remove.
         * @returns {*} Return the old value of the deleted attribute or undefined if the attribute does'nt exist.
         */
        this.removeAttr = function (attrName) {
            if (_self.hasAttr(attrName)) {
                var oldValue = _attributes[attrName];
                delete _attributes[attrName];
                return oldValue;
            } else {
                return undefined;
            }
        };
        /**
         * @description Tell if a specific attribute exist (!== undefined).
         * @param {string} attrName The name of the attribute to test.
         * @returns {boolean}
         */
        this.hasAttr = function (attrName) { return _attributes[attrName] !== undefined; };

        /**
         * @description Set a specific property.
         * @param {string} propName The name of the property to set.
         * @param {*} propValue The value to set to the property.
         * @returns {*} The value of the property.
         */
        this.setProp = function (propName, propValue) {
            _properties[propName] = propValue;
            return propValue;
        };
        /**
         * @description Get the value of a specific property.
         * @param {string} propName The name of the property to get.
         * @returns {*}
         */
        this.getProp = function (propName) { return _properties[propName]; };
        /**
         * @description Remove a specific value.
         * @param {string} propName The name of the property to remove.
         * @returns {*} Return the old value of the deleted property or undefined if the property does'nt exist.
         */
        this.removeProp = function (propName) {
            if (_self.hasProp(propName)) {
                var oldValue = _properties[propName];
                delete _properties[propName];
                return oldValue;
            } else {
                return undefined;
            }
        };
        /**
         * @description Tell if a specific property exist (!== undefined).
         * @param {string} propName The name of the property to test.
         * @returns {boolean}
         */
        this.hasProp = function (propName) { return _properties[propName] !== undefined; };

        /**
         * @description Add the specified class to this Element.
         * @param {string} className The name of the class to add.
         * @returns {boolean} Return if the job was successfully done. False means that this Element already have this class.
         */
        this.addClass = function (className) {
            if (!_self.hasClass(className)) {
                _classes.push(className);
                return true;
            } else {
                return false;
            }
        };
        /**
         * @description Remove the specified class from this Element.
         * @param {string} className The class to remove from this Element.
         * @returns {boolean} Return if the job was successfully done. False means that this Element does'nt have the specified class.
         */
        this.removeClass = function (className) {
            if (_self.hasClass(className)) {
                _classes.splice(_classes.indexOf(className), 1);
                return true;
            } else {
                return false;
            }
        };
        /**
         * @description Test if this Element has the specified class.
         * @param {string} className The name of the class to test.
         * @returns {boolean}
         */
        this.hasClass = function (className) {
            return _classes.indexOf(className) !== -1;
        };
        /**
         * @description Give all of this Element classes.
         * @returns {string[]}
         */
        this.getClasses = function () { return _classes; };
        /**
         * @description Remove all classes from this Element.
         * @returns {boolean} Return if the job was successfully done. False means that this Element does'nt have any class.
         */
        this.eraseClasses = function () {
            if (_classes.length > 0) {
                _classes = [];
                return true;
            } else {
                return false;
            }
        };

        this.data(node_a.data.hiddenProperties, {me: this, classes: _classes});
    }

    /**
     * @description Hey Element, have you 'attrName'? Similar to hasAttr.
     * @param {string} attrName
     * @returns {boolean} Return true only if the attribute is a truthy value.
     */
    Element.prototype.haveYou = function (attrName) {
        return this.hasAttr(attrName) && (this.getAttr(attrName) ? true : false);
    };
    /**
     * @description Hey Element, are you 'propName'? Similar to hasProp.
     * @param {string} propName
     * @returns {boolean} Return true only if the property is a truthy value.
     */
    Element.prototype.areYou = function (propName)  {
        return this.hasProp(propName) && (this.getProp(propName) ? true: false);
    };

    /**
     * @description A GNode Element.
     * @param {object} data
     * @constructor
     */
    function GNode (data) {
        /**
         * @description This GNode.
         * @type {GNode}
         * @private
         */
        var _self = this;
        Element.call(this, data);

        /**
         * @description Select me.
         */
        this.select = function () {
            _self.setProp(node_p.impacted, _self.setProp(node_p.selected, true));
            _self.data(node_a.data.impactNode, parseInt(myImpactLevel));
        };
        /**
         * @description Unselect me.
         */
        this.unselect = function () {
            _self.removeProp(node_p.impacted);
            _self.removeProp(node_p.selected);
            delete _self.data()[node_a.data.impactNode];
        };
        /**
         * @description Am I selected?
         * @returns {boolean}
         */
        this.selected = function () {
            return _self.areYou(node_p.selected);
        };

        /**
         * @description Hey node, show me what you are!
         * @returns {string}
         */
        this.show = function () {
            var response = "<ul><li><h3>Id</h3><p>" + this.id() + "</p></li>";

            if (debug) {
                for (var index in node_p) {
                    if (this.areYou(node_p[index])) {
                        switch (node_p[index]) {
                            case node_p.me:
                            case node_p.cytoscapeNodes:
                                break;
                            case node_p.parent:
                                response += "<li><h3>Has parents</h3>";
                                if (this.getProp(node_p.parent)[utilisation.cause]) {
                                    response += "<p>When grouped by cause: " + this.getProp(node_p.parent)[utilisation.cause].id() + "</p></li>";
                                }
                                if (this.getProp(node_p.parent)[utilisation.impact]) {
                                    response += "<p>When grouped by impact: " + this.getProp(node_p.parent)[utilisation.impact].id() + "</p></li>";
                                }
                                break;
                            case node_p.nodes:
                                response += "<li><h3>Is a group node</h3><p>With " + this.getProp(node_p.nodes).length + " nodes in.</p></li>";
                                break;
                            case node_p.precedingEdges:
                                response += "<li><h3>Has preceding edges</h3>";
                                var precedingEdges = this.precedingEdges();
                                response += "<p><table style='border=1px black solid;'>" +
                                    "<tr><th>Edge source</th><th>Edge target</th><th>Criticality</th></tr>";
                                for (var pe = 0, precedingEdge = precedingEdges[pe]; pe < precedingEdges.length; pe++, precedingEdge = precedingEdges[pe]) {
                                    response += "<tr><td>" + precedingEdge.source().id() + "</td><td>" + precedingEdge.target().id() + "</td><td>" + precedingEdge.data(edge_a.data.criticality) + "</td></tr>";
                                }
                                response += "</table></p></li>";
                                break;
                            case node_p.succeedingEdges:
                                response += "<li><h3>Has " + this.succeedingEdges().length + " succeeding edge(s).</h3>";
                                break;
                            case node_p.impacted:
                                response += "<li><h3>Is impacted</h3>" +
                                    "<p>With an impact of: " + (this.data(node_a.data.impactNode) || this.data(node_a.data.impactNode) === 0 ? this.data(node_a.data.impactNode)  + "%" : this.data(node_a.data.impactNode)) + "</p></li>";
                                break;
                            case node_p.causeOf:
                                var cause = this.getProp(node_p.causeOf);
                                response += "<li><h3>Is cause of:</h3>";
                                if (cause.length > 1) {
                                    response += "<p>1 node so it is just a simple cause.</p></li>";
                                } else {
                                    response += "<p>" + cause.length + " nodes so it is an intersection.</p></li>";
                                }
                                break;
                            case node_p.filter:
                                response += "<li><h3>Is Filtered</h3></li>";
                                break;
                            default:
                                response += "<li><h3>" + index.substring(0,1).toUpperCase() + index.substring(1).toLowerCase() + "</h3>" +
                                    "<p>" + this.getProp(node_p[index]) + "</p></li>";
                                break;
                        }
                    }
                }
                var data = this.data();
                for (var index in data) {
                    if (index.search(/^nb_/) !== -1) {
                        response += "<li><h3>" + index.substring(0,1).toUpperCase() + index.substring(1).toLowerCase() + "</h3>" +
                            "<p>" + data[index] + "</p></li>";
                    } else if (index === "redundant") {
                        response += "<li><h3>Redundant</h3><p>" + data[index] + "</p></li>";
                    }
                }
                response += "<li><h3>Classes</h3><p>" + this.getClasses().join(" ") + "</p></li>";
            } else {
                var data = this.data();
                for (var index in data) {
                    switch (index) {
                        case node_a.data.id:
                        case node_a.data.hiddenProperties:
                        case node_a.data.class:
                        case node_a.data.description:
                        case node_a.data.parent:
                            break;
                        default:
                            response += "<li><h3>" + index.substring(0,1).toUpperCase() + index.substring(1).toLowerCase() + "</h3>" +
                                "<p>" + data[index] + "</p></li>";
                            break;
                    }
                }
            }

            return response + "</ul>";
        };
    }
    GNode.prototype = Object.create(Element.prototype);
    GNode.prototype.constructor = GNode;

    /**
     * @description Give the preceding GEdge of this GNode.
     * @returns {GEdge[]}
     */
    GNode.prototype.precedingEdges = function () { return this.getProp(node_p.precedingEdges); };
    /**
     * @description Give the succeeding GEdge of this GNode.
     * @returns {GEdge[]}
     */
    GNode.prototype.succeedingEdges = function () { return this.getProp(node_p.succeedingEdges); };
    /**
     * @description Give the preceding GNode of this GNode.
     * @returns {GNode[]}
     */
    GNode.prototype.precedingNodes = function () {
        var precedingNodes = [];
        for (var precedingEdges = this.precedingEdges(), se = 0, precedingEdge = precedingEdges[se];
             se < precedingEdges.length;
             se++, precedingEdge = precedingEdges[se])
        {
            precedingNodes.push(precedingEdge.source());
        }
        return precedingNodes;
    };
    /**
     * @description Give the succeeding GNode of this GNode.
     * @returns {GNode[]}
     */
    GNode.prototype.succeedingNodes = function () {
        var succeedingNodes = [];
        for (var succeedingEdges = this.succeedingEdges(), se = 0, succeedingEdge = succeedingEdges[se];
                                    se < succeedingEdges.length;
                            se++, succeedingEdge = succeedingEdges[se])
        {
            succeedingNodes.push(succeedingEdge.target());
        }
        return succeedingNodes;
    };

    /**
     * @description Tell if this GNode has predecessors.
     * @returns {boolean}
     */
    GNode.prototype.hasPredecessors = function () { return this.hasProp(node_p.precedingEdges) && this.getProp(node_p.precedingEdges).length > 0; };
    /**
     * @description Tell if this GNode has successors.
     * @returns {boolean}
     */
    GNode.prototype.hasSuccessors = function () { return this.hasProp(node_p.succeedingEdges) && this.getProp(node_p.succeedingEdges).length > 0; };

    /**
     * @description A GEdge Element.
     * @param {object} data
     * @constructor
     */
    function GEdge (data) {
        Element.call(this, data);

        /**
         * @description Hey edge, show me what you are!
         * @returns {string}
         */
        this.show = function () {
            var response = "<ul>";
            if (debug) {
                for (var index in edge_p) {
                    response += "<li>";
                    switch (edge_p[index]) {
                        case edge_p.me:
                        case edge_p.moving:
                            break;
                        case edge_p.source:
                            response += "<h3>Source</h3><p>" + this.source().id() + "</p>";
                            break;
                        case edge_p.target:
                            response += "<h3>Target</h3><p>" + this.target().id() + "</p>";
                            break;
                        case edge_p.nodes:
                            response += "<h3>Filter nodes</h3><p>" + this.getProp(edge_p.nodes).length + " node(s).</p>";
                            break;
                    }
                    response += "</li>";
                }
            } else {
                /* TODO: AFFICHER TOUTES LES DATAS. */
                var data = this.data();
                for (var index in data) {
                    response += "<li>";
                    switch (index) {
                        case edge_a.data.hiddenProperties:
                        case edge_a.data.id:
                        case edge_a.data.realSource:
                        case edge_a.data.realTarget:
                            break;
                        default:
                            response += "<h3>" + index.substring(0,1).toUpperCase() + index.substring(1).toLowerCase() + "</h3>" +
                                "<p>" + this.data(index) + "</p>";
                            break;
                    }
                    response += "</li>";
                }
            }
            return response + "</ul>";
        }
    }
    GEdge.prototype = Object.create(Element.prototype);
    GEdge.prototype.constructor = GEdge;

    /**
     * @description Give the GNode source of this GEdge.
     * @returns {GNode}
     */
    GEdge.prototype.source = function () { return this.getProp(edge_p.source); };
    /**
     * @description Give the GNode target of this GEdge.
     * @returns {GNode}
     */
    GEdge.prototype.target = function () { return this.getProp(edge_p.target); };

    return {GNode: GNode, GEdge: GEdge};
    // Comment-ajout: on aurais pu faire
    // return [GNode, GEdge];
    // Il aurait fallu récupérer les éléments comme suit : 
    // [variable1, variable2] = (function () { return [valeur1, valeur2]; } ());
    // C'est une petite facilité de JavaScript qui permet de retourné plusieurs éléments dans un tableau et de les récupérer dans plusieurs variables différentes.
}());

GNode = getElements.GNode;
GEdge = getElements.GEdge;
getElements = undefined; // Ici je supprime la référence sur l'objet pour forcer l'utilisation des objets GNode et GEdge.